@@ -176,7 +176,7 @@ Rcpp::NumericVector eval_log_lm_terms(
   openmp_exception_ptr capture_err;
 
 #ifdef _OPENMP
-#pragma omp parallel
+#pragma omp parallel num_threads(n_threads)
 #endif
   {
     arma::mat my_derivs_vcov(comp_derivs ? p : 0L, comp_derivs ? p : 0L,
@@ -224,7 +224,7 @@ Rcpp::NumericVector eval_log_lm_terms(
 // [[Rcpp::export(rng = false)]]
 Rcpp::NumericMatrix get_z_hat
 (arma::mat const &lower, arma::mat const &upper, arma::imat const &code,
- unsigned const n_threads, Rcpp::List multinomial){
+ int const n_threads, Rcpp::List multinomial){
   size_t const p = lower.n_rows,
                n = upper.n_cols;
   if(upper.n_rows != p or upper.n_cols != n)
@@ -244,6 +244,10 @@ Rcpp::NumericMatrix get_z_hat
       break;
     }
 
+#ifdef _OPENMP
+  omp_set_num_threads(n_threads);
+#endif
+
   Rcpp::NumericMatrix out(p, n);
   double * const o = &out[0];
   openmp_exception_ptr capture_err;
@@ -690,7 +694,7 @@ Rcpp::List impute
   openmp_exception_ptr capture_err;
 
 #ifdef _OPENMP
-#pragma omp parallel for schedule(static) \
+#pragma omp parallel for schedule(static) num_threads(n_threads) \
   firstprivate(w_idx_int, w_idx_obs, w_obs_val, w_upper, w_lower, \
                mu, w_idx_cat_obs, w_idx_cat_not_obs, known_objs) \
   private(a_idx_int, a_idx_obs, a_obs_val, a_upper, a_lower, \
@@ -969,9 +973,10 @@ Rcpp::NumericVector lower_tri_inner
   if(idx.nrow() < 1)
     return Rcpp::NumericVector();
 
-  double const fdim = .5 * (std::sqrt(8 * x.size() + 1) - 1);
+  double const fdim = .5 * (std::sqrt(
+    8. * static_cast<double>(x.size()) + 1.) - 1.);
   int const dim = std::lround(fdim);
-  if(std::abs(fdim / dim - 1) >
+  if(std::fabs(fdim / static_cast<double>(dim) - 1) >
        std::numeric_limits<double>::epsilon() * 10)
     throw std::invalid_argument("lower_tri_outer: invalid x");
   if(idx.ncol() != 2L)
@@ -1047,7 +1052,7 @@ Rcpp::NumericVector multinomial_find_means
   (arma::vec const &probs, double const rel_eps = 3.000214e-13,
    int const max_it = 100, double const c1 = .0001,
    double const c2 = .9){
-  if(probs.size() < 2 or std::abs(arma::sum(probs) - 1) >= 1e-10)
+  if(probs.size() < 2 or std::fabs(arma::sum(probs) - 1) >= 1e-10)
     throw std::invalid_argument("multinomial_find_means: invalid probs");
 
   Rcpp::NumericVector mu(probs.size() - 1);
